@@ -372,12 +372,31 @@ static int setup_opts_from_req(int sk, CriuOpts *req)
 			goto err;
 	}
 
+	for (i = 0; i < req->n_enable_fs; i++) {
+		if (!add_fsname_auto(req->enable_fs[i]))
+			goto err;
+	}
+
+	for (i = 0; i < req->n_skip_mnt; i++) {
+		if (!add_skip_mount(req->skip_mnt[i]))
+			goto err;
+	}
+
 	if (req->has_cpu_cap)
 		opts.cpu_cap = req->cpu_cap;
 
 	if (req->has_manage_cgroups)
 		opts.manage_cgroups = req->manage_cgroups;
 
+	if (req->has_auto_ext_mnt)
+		opts.autodetect_ext_mounts = req->auto_ext_mnt;
+
+	if (req->has_ext_sharing)
+		opts.enable_external_sharing = req->ext_sharing;
+
+	if (req->has_ext_masters)
+		opts.enable_external_masters = req->ext_masters;
+
 	return 0;
 
 err:
