@@ -57,6 +57,10 @@
 #include "protobuf/sk-netlink.pb-c.h"
 #include "protobuf/vma.pb-c.h"
 #include "protobuf/tun.pb-c.h"
+#include "protobuf/cgroup.pb-c.h"
+#include "protobuf/timerfd.pb-c.h"
+#include "protobuf/cpuinfo.pb-c.h"
+#include "protobuf/userns.pb-c.h"
 
 struct cr_pb_message_desc cr_pb_descs[PB_MAX];
 
